/***********************************************************************
 * Copyright (c) 2007, Devin Papineau <jdmpapin@uwaterloo.ca>
 ***********************************************************************
 *
 *  This software is provided 'as-is', without any express or implied
 *  warranty.  In no event will the authors be held liable for any
 *  damages arising from the use of this software.
 *
 *  Permission is granted to anyone to use this software for any
 *  purpose, including commercial applications, and to alter it and
 *  redistribute it freely, subject to the following restrictions:
 *
 *  1. The origin of this software must not be misrepresented; you
 *     must not claim that you wrote the original software. If you use
 *     this software in a product, an acknowledgment in the product
 *     documentation would be appreciated but is not required.
 *
 *  2. Altered source versions must be plainly marked as such, and
 *     must not be misrepresented as being the original software.
 *
 *  3. This notice may not be removed or altered from any source
 *     distribution.
 *
 ***********************************************************************
 *
 *  This program generates an image of a region of the Mandelbrot set.
 *
 *  It requires the DevIL library
 *
 *      http://openil.sourceforge.net/download.php
 *
 *  and posix threads. Link with -lm -lIL -lILU -lpthread.
 *
 ***********************************************************************
 *
 *  - If a gradient file is not specified, this should be run from the
 *    directory containing default_gradient.
 *
 *  - TTY width isn't currently detected properly.
 *
 *  - getopt_long and pthreads probably aren't too Windows-friendly.
 *
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <getopt.h>
#include <errno.h>
#include <IL/il.h>
#include <IL/ilu.h>
#include <pthread.h>

#define X_MIN_DEFAULT (-2.0)
#define X_MAX_DEFAULT 1.0
#define Y_MIN_DEFAULT (-1.0)
#define Y_MAX_DEFAULT 1.0
#define X_RES_DEFAULT 450
#define Y_RES_DEFAULT 300
#define ITERATIONS_DEFAULT (1<<16)
#define ESCAPE_DEFAULT 1.0e128
#define CDIFF_DEFAULT 5.0
#define CDIFF_SCALE_DEFAULT 1.25
#define MULTISAMPLE_DEFAULT 1
#define OUT_DEFAULT "out.png"
#define GRADIENT_DEFAULT "default_gradient"
#define JOBS_DEFAULT 1

/* Global parameter structure. Yeah, yeah, I know, bad idea. Whatever. */
struct parameters {
    double x_min, x_max;
    double y_min, y_max;
    int x_res, y_res;
    double pix_w, pix_h;
    double subpix_w, subpix_h;
    int iterations;
    double escape; /* actually the square of the escape value */
    double reciprocal_square_multisample; /* multiply instead of divide :D */
    char *out_filename;
    double cdiff;
    double cdiff_scale;
    int multisample;
    char *gradient_file;
    int n_colors;
    float *gradient; // has n_colors elements
    int progress_bar;
    int term_width;
    int verbose;

    int jobs; // parallelism :D
} params;

struct accounting {
    pthread_mutex_t lock;
    int row, col;
    char spinner_chars[4];
    int spinner;
    float progress;
    char *progress_bar;
    int bar_length;
    unsigned char *image;
} accounting;

void usage(FILE *out, int retcode, const char *self)
{
    fprintf(out,
            "Usage: %s [OPTION]...\n"
            "Generate a Mandelbrot fractal image.\n"
            "Example: %s -v -o fractal.png -i 1000\n"
            "\n"
            "Miscellaneous:\n"
            "  -h, --help       display this help and exit\n"
            "  -v, --verbose    be more verbose\n"
            "  -p, --progress   display a textual progress bar\n"
            "\n"
            "Complex window co-ordinates:\n"
            "  -x, --domain=MIN,MAX min,max real values [-2,1]\n"
            "  -y, --range=MIN,MAX  min,max imaginary values [-1,1]\n"
            "  -c, --center=RE,IM   window center\n"
            "  -d, --diameter=D     horizontal (real) diameter. vertical\n"
            "                       is proportional according to the output resolution\n"
            "  NOTE: if -x or -y is used, then the other must also be used, and\n"
            "        -c and -d cannot be used, and vice versa\n"
            "\n"
            "Generation:\n"
            "  -i, --iterations=I   iterate I times [%d]\n"
            "  -e, --escape=E       the square of the escape modulus value [1e128]\n"
            "  -j, --jobs=JOBS      compute the colour for JOBS pixels in parallel\n"
            "\n"
            "Coloring:\n"
            "  -D, --cdiff=DIFF     starting difference between gradient colors [5]\n"
            "  -S, --cdiff-scale=S  scale DIFF by S at every gradient color [1.25]\n"
            "  -G, --gradient=FILE  read gradient from FILE [default_gradient]\n"
            "  -m, --multisample=N  sample NxN `sub-pixels' and average [1]\n"
            "\n"
            "Output:\n"
            "  -o, --out=FILENAME   output to FILENAME. image type determined by extension.\n"
            "                       currently only PNG is supported [%s]\n"
            "  -r, --resolution=WxH resolution of the output image [%dx%d]\n",

            self, self,
            ITERATIONS_DEFAULT,
            OUT_DEFAULT,
            X_RES_DEFAULT, Y_RES_DEFAULT);

    exit(retcode);
}

/* Helpful I/O stuff. */
int parse_int(const char *text, int *ret)
{
    char *endptr;
    int result = strtol(text, &endptr, 10);

    if(*endptr) {
        return -1;
    }

    *ret = result;
    return 0;
}

int parse_double(const char *text, double *ret)
{
    char *endptr;
    double result = strtod(text, &endptr);

    if(*endptr) {
        return -1;
    }

    *ret = result;
    return 0;
}

int parse_delimited_pair(const char *text, char delim, double *pa, double *pb)
{
    char *c, *s = strdup(text);

    for(c = s; *c && *c != delim; c++);

    if(*c != delim) {
        free(s);
        return -1;
    }

    *c = '\0';
    c++;

    double a, b;

    if(parse_double(s, &a)) {
        fprintf(stderr, "bad number format: %s\n", s);
        free(s);
        return -1;
    }

    if(parse_double(c, &b)) {
        fprintf(stderr, "bad number format: %s\n", c);
        free(s);
        return -1;
    }

    *pa = a;
    *pb = b;

    free(s);
    return 0;
}

/**
 * Read gradient from gradient_file.
 * Each non-blank line defines a color, in sequence.
 * A color is three whitespace-separated floating point numbers (RGB),
 * where 1 1 1 is white and 0 0 0 is black.
 *
 * I'm lazy, so lines which are not in the appropriate format are ignored;
 * so is everything after the third floating point number on a line which
 * does define a colour (in theory, haven't bothered testing that). Hey,
 * I'm just trying to document here...
 */
void read_gradient_file()
{
    FILE *fp = fopen(params.gradient_file, "r");
    if(!fp) {
        fprintf(stderr, "%s: %s\n", params.gradient_file, strerror(errno));
        exit(1);
    }

    params.gradient = NULL;
    params.n_colors = 0;

    float floatbuf[3];
    char buffer[2048];

    while(!feof(fp)) {
        fgets(buffer, sizeof(buffer), fp);
        if(buffer[strlen(buffer)-1] == '\n') {
            // get rid of newline
            buffer[strlen(buffer)-1] = '\0';
        }

        int i;
        for(i = 0; i < strlen(buffer); i++) {
            if(buffer[i] != ' ' && buffer[i] != '\t') { break; }
        }
        if(i == strlen(buffer)) { continue; }

        if(sscanf(buffer,
                  "%f %f %f",
                  &floatbuf[0],
                  &floatbuf[1],
                  &floatbuf[2]) < 3) {
            // ignore, sure
            continue;
        }

        params.n_colors++;

        float *new_gradient = realloc(params.gradient,
                                      params.n_colors * sizeof(float) * 3);
        if(!new_gradient) {
            fprintf(stderr, "out of memory\n");
            exit(1);
        }

        new_gradient[(params.n_colors-1)*3+0] = floatbuf[0];
        new_gradient[(params.n_colors-1)*3+1] = floatbuf[1];
        new_gradient[(params.n_colors-1)*3+2] = floatbuf[2];

        params.gradient = new_gradient;
    }

    if(params.n_colors == 0) {
        fprintf(stderr, "%s: no colors specified\n", params.gradient_file);
        exit(1);
    }

    fclose(fp);
}

/**
 * Use getopt_long to parse command line parameters, and calculate certain
 * static info based on them (also held in params).
 */
void get_parameters(int argc, char **argv)
{
    double xres, yres;

    /* defaults */
    params.x_min = X_MIN_DEFAULT;
    params.x_max = X_MAX_DEFAULT;
    params.y_min = Y_MIN_DEFAULT;
    params.y_max = Y_MAX_DEFAULT;
    params.x_res = X_RES_DEFAULT;
    params.y_res = Y_RES_DEFAULT;
    params.iterations = ITERATIONS_DEFAULT;
    params.escape = ESCAPE_DEFAULT;
    params.cdiff = CDIFF_DEFAULT;
    params.cdiff_scale = CDIFF_SCALE_DEFAULT;
    params.gradient_file = GRADIENT_DEFAULT;
    params.multisample = MULTISAMPLE_DEFAULT;
    params.out_filename = OUT_DEFAULT;
    params.progress_bar = 0;
    params.verbose = 0;

    params.jobs = JOBS_DEFAULT;

    static struct option long_options[] = {
        { "help", no_argument, NULL, 'h' },
        { "verbose", no_argument, NULL, 'v' },
        { "progress", no_argument, NULL, 'p' },
        { "domain", required_argument, NULL, 'x' },
        { "range", required_argument, NULL, 'y' },
        { "center", required_argument, NULL, 'c' },
        { "diameter", required_argument, NULL, 'd' },
        { "resolution", required_argument, NULL, 'r' },
        { "iterations", required_argument, NULL, 'i' },
        { "escape", required_argument, NULL, 'e' },
        { "cdiff", required_argument, NULL, 'D' },
        { "cdiff-scale", required_argument, NULL, 'S' },
        { "gradient", required_argument, NULL, 'G' },
        { "multisample", required_argument, NULL, 'm' },
        { "out", required_argument, NULL, 'o' },
        { "jobs", required_argument, NULL, 'j' },
        { 0, 0, 0, 0 }
    };

    int using_center = 0;
    int using_diameter = 0;
    int using_domain = 0;
    int using_range = 0;

    double center_x, center_y, diameter;

    int opt;
    while((opt=getopt_long(argc,
                           argv,
                           "hvpx:y:c:d:r:i:e:D:S:m:o:G:j:",
                           long_options,
                           NULL)) != -1) {
        switch(opt) {
        case 'h':
            usage(stdout, 0, argv[0]);
            break;

        case 'v':
            params.verbose = 1;
            break;

        case 'p':
            params.progress_bar = 1;
            break;

        case 'x':
            if(parse_delimited_pair(optarg,
                                    ',',
                                    &params.x_min,
                                    &params.x_max)) {
                usage(stderr, 1, argv[0]);
            }
            using_domain = 1;
            break;

        case 'y':
            if(parse_delimited_pair(optarg,
                                    ',',
                                    &params.y_min,
                                    &params.y_max)) {
                usage(stderr, 1, argv[0]);
            }
            using_range = 1;
            break;

        case 'c':
            if(parse_delimited_pair(optarg,
                                    ',',
                                    &center_x,
                                    &center_y)) {
                usage(stderr, 1, argv[0]);
            }
            using_center = 1;
            break;

        case 'd':
            if(parse_double(optarg, &diameter)) {
                fprintf(stderr, "bad number format: %s\n", optarg);
                usage(stderr, 1, argv[0]);
            }
            using_diameter = 1;
            break;

        case 'r':
            if(parse_delimited_pair(optarg,
                                    'x',
                                    &xres,
                                    &yres)) {
                usage(stderr, 1, argv[0]);
            }
            params.x_res = (int)xres;
            params.y_res = (int)yres;
            break;

        case 'i':
            if(parse_int(optarg, &params.iterations)) {
                fprintf(stderr, "invalid integer format: %s\n", optarg);
                usage(stderr, 1, argv[0]);
            }
            break;

        case 'e':
            if(parse_double(optarg, &params.escape)) {
                fprintf(stderr, "invalid number format: %s\n", optarg);
                usage(stderr, 1, argv[0]);
            }
            break;

        case 'D':
            if(parse_double(optarg, &params.cdiff)) {
                fprintf(stderr, "invalid number format: %s\n", optarg);
                usage(stderr, 1, argv[0]);
            }
            break;

        case 'S':
            if(parse_double(optarg, &params.cdiff_scale)) {
                fprintf(stderr, "invalid number format: %s\n", optarg);
                usage(stderr, 1, argv[0]);
            }
            break;

        case 'm':
            if(parse_int(optarg, &params.multisample)) {
                fprintf(stderr, "invalid integer format: %s\n", optarg);
                usage(stderr, 1, argv[0]);
            }
            break;

        case 'G':
            params.gradient_file = strdup(optarg);
            break;

        case 'o':
            params.out_filename = strdup(optarg);
            break;

        case 'j':
            params.jobs = atoi(optarg);
            if(params.jobs <= 0) {
                fprintf(stderr, "nonpositive number of jobs\n");
                exit(1);
            }
            break;

        case '?':
            usage(stderr, 1, argv[0]);
            break;
        }
    }

    /* We can't use both domain/range and center/diameter.
     * If neither is in use, we don't care, just use default.
     * Otherwise we have some checking to do. */
    if(using_domain || using_range || using_center || using_diameter) {
        if(using_center && !using_diameter) {
            fprintf(stderr, "center specified without diameter\n");
            usage(stderr, 1, argv[0]);
        } else if(using_diameter && !using_center) {
            fprintf(stderr, "diameter specified without center\n");
            usage(stderr, 1, argv[0]);
        }

        if(using_domain && !using_range) {
            fprintf(stderr, "domain specified without range\n");
            usage(stderr, 1, argv[0]);
        } else if(using_range && !using_domain) {
            fprintf(stderr, "range specified without domain\n");
            usage(stderr, 1, argv[0]);
        }

        if(using_center && using_domain) {
            fprintf(stderr, "both domain/range and center/diameter specified\n");
            usage(stderr, 1, argv[0]);
        }

        if(using_center) {
            double im_diam = diameter*(double)params.y_res/(double)params.x_res;
            double re_r = diameter /= 2.0;
            double im_r = im_diam /= 2.0;

            params.x_min = center_x - re_r;
            params.x_max = center_x + re_r;
            params.y_min = center_y - im_r;
            params.y_max = center_y + im_r;
        }
    }

    params.pix_w = (params.x_max - params.x_min) / params.x_res;
    params.pix_h = (params.y_max - params.y_min) / params.y_res;
    params.subpix_w = params.pix_w / (params.multisample+1);
    params.subpix_h = params.pix_h / (params.multisample+1);
    params.reciprocal_square_multisample =
        1.0 / (params.multisample * params.multisample);

    /* This isn't right, and I ought to use curses or something. */
    params.term_width = 80;

    read_gradient_file();
}

/**
 * Colouring is done by a sliding-scale circular gradient using
 * smoothed fractional iteration numbers. The cdiff parameter
 * specifies the initial difference in iterations between subsequent
 * colours. When a colour is passed, the difference is scaled by the
 * cscale parameter. Trust me, I tried everything. This is the only
 * thing that wasn't ugly as hell.
 */
void gradient(double n, double *r, double *g, double *b)
{
    double diff = params.cdiff;;
    int color, color1;

    float *colors = &params.gradient[0];

    for(color = 0; n > diff; n -= diff, color++, diff *= params.cdiff_scale);
    color %= params.n_colors;
    color1 = (color+1) % params.n_colors;

    /* LERP with t=n/diff. */
    n /= diff;
    *r = colors[color*3+0] + n * (colors[color1*3+0] - colors[color*3+0]);
    *g = colors[color*3+1] + n * (colors[color1*3+1] - colors[color*3+1]);
    *b = colors[color*3+2] + n * (colors[color1*3+2] - colors[color*3+2]);
}

/**
 * Determine if a complex number a + bi is in the Mandelbrot set.
 * Return -1 if it's in the set, and otherwise return a smoothed
 * iteration count required to escape.
 */
double mandelbrot(double a, double b)
{
    double a0=a, b0=b;

    int i;
    double aa, bb;
    for(i = 0; i < params.iterations; i++) {
        aa = a*a;
        bb = b*b;
        if(aa + bb > params.escape) {
            /* I found this smoothing expression on some site. */
            return i + 1.0 - log(log(sqrt(aa+bb))) / log(2);
        }
        b *= 2*a;
        b += b0;
        a = aa - bb + a0;
    }

    return -1.0;
}

/**
 * Sample multisample X multisample `sub-points' of the pixel at x,y
 * relative to top-left of image, average, and write into rgb.
 * rgb[0] == r
 * rgb[1] == g
 * rgb[2] == b
 */
void color_pixel(int x, int y, unsigned char *rgb)
{
    double re = params.x_min + params.pix_w * x;
    double im = params.y_min + params.pix_h * y;

    double r, g, b;
    double r_avg=0.0, g_avg=0.0, b_avg=0.0;
    double n;

    int i,j;
    for(i = 0; i < params.multisample; i++) {
        for(j = 0; j < params.multisample; j++) {
            n = mandelbrot(re + i * params.subpix_w, im - j * params.subpix_h);
            if(n >= 0.0) {
                gradient(n, &r, &g, &b);
                r_avg += r;
                g_avg += g;
                b_avg += b;
            }
        }
    }

    /* Finish averaging. Then scale into unsigned char range. No
     * truncation to 0 for you!  Come back, one year. */
    rgb[0] = 255 * r_avg * params.reciprocal_square_multisample;
    rgb[1] = 255 * g_avg * params.reciprocal_square_multisample;
    rgb[2] = 255 * b_avg * params.reciprocal_square_multisample;
}

/**
 * Use DevIL to write to out_filename.
 */
int write_image()
{
    ILuint img;
    ilGenImages(1, &img);

    if(!img) {
        fprintf(stderr, "%s\n", iluErrorString(ilGetError()));
        return -1;
    }

    ilBindImage(img);

    ILenum err;

    ilTexImage(params.x_res,
               params.y_res,
               1,
               3,
               IL_RGB,
               IL_UNSIGNED_BYTE,
               accounting.image);

    err = ilGetError();
    if(err != IL_NO_ERROR) {
        fprintf(stderr, "%s\n", iluErrorString(err));
        ilDeleteImages(1, &img);
        return -1;
    }

    ilEnable(IL_FILE_OVERWRITE);
    ilSaveImage(params.out_filename);
    err = ilGetError();
    if(err != IL_NO_ERROR) {
        fprintf(stderr, "%s: %s\n", params.out_filename, iluErrorString(err));
        ilDeleteImages(1, &img);
        return -1;
    }

    ilDeleteImages(1, &img);
    return 0;
}

/**
 * Draw the blasted progress bar.
 */
void draw_progress()
{
    if(params.progress_bar) {
        fprintf(stderr,
                "\r%c %3d%%[%s]",
                accounting.spinner_chars[accounting.spinner],
                (int)(100*accounting.progress),
                accounting.progress_bar);
        fflush(stdout);
    }
}

/**
 * Clean up resources.
 */
void cleanup()
{
    pthread_mutex_destroy(&accounting.lock);
    free(accounting.image);
    free(params.gradient);
}

/**
 * This is a thread.
 */
void *worker_function(void *arg)
{
    int i, j;
    int ret;

    while(1) {
        // Grab the lock to take the next set of coordinates.
        // only printf while we have the lock in order to prevent issues.
        {
            // This scope visually separates what is done with the lock held.
            if((ret = pthread_mutex_lock(&accounting.lock))) {
                // can't happen
                fprintf(stderr, "pthread_mutex_lock: %s\n", strerror(ret));
                exit(1); // cleanup() will fail here, so just quit, dammit
            }

            j = accounting.col++;
            i = accounting.row;

            if(accounting.col >= params.x_res) {
                accounting.row++;
                accounting.col = 0;
            }

            if(j + 1 == params.x_res && params.progress_bar) {
                accounting.progress = (float)(i+1) / (float)params.y_res;

                accounting.spinner++;
                accounting.spinner %= 4; // 4 spinner chars

                int n_bar_chars = accounting.progress * accounting.bar_length;
                int k;
                for(k = 0; k < n_bar_chars; k++) {
                    accounting.progress_bar[k] = '#';
                }

                draw_progress();
            }

            if((ret = pthread_mutex_unlock(&accounting.lock))) {
                // also can't happen
                fprintf(stderr, "pthread_mutex_unlock: %s\n", strerror(ret));
                exit(1); // same deal
            }
        }

        if(i >= params.y_res) {
            pthread_exit(NULL);
        }

        // This is the bit that takes a while.
        // It can't step on other threads' turf because we worked out
        // the coordinates to be unique.
        color_pixel(j, i, &accounting.image[3 * (j + i*params.x_res)]);
    }
}

/**
 * Put it all together.
 */
int main(int argc, char **argv)
{
    ilInit();
    iluInit();

    get_parameters(argc, argv);

    if(params.verbose) {
        fprintf(stderr,
                "visualizing rectangle (%f, %f), (%f, %f)\n"
                "output to %s, at %dx%d\n"
                "%d max iterations\n"
                "escape value %f\n"
                "%dx multisampling\n"
                "pixel dimensions %fx%f\n"
                "initial iteration difference b/w colours %f\n"
                "difference scaling factor %f\n",
                params.x_min, params.y_min,
                params.x_max, params.y_max,
                params.out_filename, params.x_res, params.y_res,
                params.iterations,
                sqrt(params.escape),
                params.multisample,
                params.pix_w, params.pix_h,
                params.cdiff, params.cdiff_scale);
    }

    accounting.image = malloc(params.x_res * params.y_res * 3);
    if(!accounting.image) {
        perror("malloc");
        free(params.gradient);
        return -1;
    }

    accounting.spinner_chars[0] = '|';
    accounting.spinner_chars[1] = '/';
    accounting.spinner_chars[2] = '-';
    accounting.spinner_chars[3] = '\\';

    accounting.progress = 0.0;
    accounting.spinner = 0;
    accounting.bar_length = params.term_width - strlen("| 100%[]");
    accounting.progress_bar = malloc(accounting.bar_length+1);
    memset(accounting.progress_bar, ' ', accounting.bar_length);
    accounting.progress_bar[accounting.bar_length] = 0;

    if(params.verbose) {
        fprintf(stderr, "rendering\n");
    }

    accounting.row = 0;
    accounting.col = 0;

    int ret;

    if((ret = pthread_mutex_init(&accounting.lock, NULL))) {
        fprintf(stderr, "pthread_mutex_init: %s\n", strerror(ret));
        free(accounting.image);
        free(params.gradient);
        exit(1);
    }

    if(params.verbose) {
        fprintf(stderr, "creating %d worker threads\n", params.jobs);
    }

    int i;
    pthread_t *workers = malloc(sizeof(pthread_t) * params.jobs);

    if(!workers) {
        perror("malloc");
        cleanup();
        exit(1);
    }

    for(i = 0; i < params.jobs; i++) {
        if((ret = pthread_create(&workers[i], NULL, worker_function, NULL))) {
            fprintf(stderr, "pthread_create: %s\n", strerror(ret));
            cleanup();
            exit(1);
        }
    }

    void *voidret;
    for(i = 0; i < params.jobs; i++) {
        pthread_join(workers[i], &voidret);
    }

    if(params.progress_bar) {
        fprintf(stderr, "\n");
    }

    if(params.verbose) {
        fprintf(stderr, "writing to %s\n", params.out_filename);
    }

    if(write_image()) {
        cleanup();
        return 2;
    }

    if(params.verbose) {
        fprintf(stderr, "done\n");
    }

    cleanup();

    return 0;
}
